/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "HostCalls.h"
#include "GServClient.h"
#include "GServerMsgs.h"
#include "PackedArgs.h"
#include "DbgPrintf.h"
#include "SecureFuncs.h"

UINT
HostCall_CFH_Begin(
    IN uint32_t ClientId,
    IN uint32_t DropHandle,
    OUT uint32_t *pHandle
    )
{
    CPackedArgs InputArgs;

    UINT uMessage = GSM_CFH_BEGIN;

    InputArgs.CreateFromArgs2(
        CPackedArgs::ARGT_UINT32, &uMessage,
        CPackedArgs::ARGT_UINT32, &DropHandle
        );

    PVOID pInputBuffer = NULL;
    ULONG InputBufferSize = 0;

    if ( !InputArgs.Pack( &pInputBuffer, &InputBufferSize ) )
    {
        return ERROR_OUTOFMEMORY;
    }

    PVOID pOutputBuffer = NULL;
    ULONG OutputBufferSize = 0 ;

    UINT uReturnCode = CallGuestServer(
        ClientId,
        pInputBuffer,
        InputBufferSize,
        &pOutputBuffer,
        &OutputBufferSize
        );

    CPackedArgs::FreeMem( pInputBuffer );

    if ( uReturnCode != ERROR_SUCCESS )
        return uReturnCode;

    CPackedArgs OutputArgs;

    if ( !OutputArgs.Unpack( pOutputBuffer, OutputBufferSize ) )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    //u32ErrorCode, u32Handle

    if ( GSM_CFH_BEGIN_OUT_CPARAMS != OutputArgs.GetCount() )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    UINT u32ErrorCode = 0;
    UINT u32Handle = 0;

    if ( !OutputArgs.GetUInt32Arg( 0, &u32ErrorCode ) ||
         !OutputArgs.GetUInt32Arg( 1, &u32Handle )
        )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    *pHandle = u32Handle;

    GS_FreeMem( pOutputBuffer );
    return u32ErrorCode;
}

UINT
HostCall_CFH_GetBuildingListStatus(
    IN  UINT ClientId,
	IN  UINT Handle,
	OUT PCOPY_FILES_BUILDING_LIST_STATUS pCopyFilesBuildingListStatus
	)
{
    CPackedArgs InputArgs;

    UINT uMessage = GSM_CFH_GET_BUILDING_LIST_STATUS; //MessageId

    //Input params

    InputArgs.CreateFromArgs2(
        CPackedArgs::ARGT_UINT32, &uMessage,
        CPackedArgs::ARGT_UINT32, &Handle
        );

    //-----------------------

    PVOID pInputBuffer = NULL;
    ULONG InputBufferSize = 0;

    if ( !InputArgs.Pack( &pInputBuffer, &InputBufferSize ) )
    {
        return ERROR_OUTOFMEMORY;
    }

    PVOID pOutputBuffer = NULL;
    ULONG OutputBufferSize = 0 ;

    UINT uReturnCode = CallGuestServer(
        ClientId,
        pInputBuffer,
        InputBufferSize,
        &pOutputBuffer,
        &OutputBufferSize
        );

    CPackedArgs::FreeMem( pInputBuffer );

    if ( uReturnCode != ERROR_SUCCESS )
        return uReturnCode;

    CPackedArgs OutputArgs;

    if ( !OutputArgs.Unpack( pOutputBuffer, OutputBufferSize ) )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    //------------------------------

    //u32ErrorCode, PCOPY_FILES_BUILDING_LIST_STATUS 

    if ( GSM_CFH_GET_BUILDING_LIST_STATUS_OUT_CPARAMS != OutputArgs.GetCount() )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    UINT u32ErrorCode = 0;
    PVOID pPtr = NULL;
    ULONG PtrSize = 0;

    if ( !OutputArgs.GetUInt32Arg( 0, &u32ErrorCode ) ||
         !OutputArgs.GetPtrArg( 1, &pPtr, &PtrSize ) ||
         PtrSize != sizeof(COPY_FILES_BUILDING_LIST_STATUS)
        )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    RtlCopyMemory( pCopyFilesBuildingListStatus, pPtr, sizeof(COPY_FILES_BUILDING_LIST_STATUS) );

    GS_FreeMem( pOutputBuffer );
    return u32ErrorCode;

}

UINT
HostCall_CFH_GetCurrentFileName(
    IN  UINT ClientId,
	IN  UINT Handle,
	OUT PCOPY_FILES_GET_CURRENT_FILENAME pCopyFilesGetCurrentFileName
	)
{
    CPackedArgs InputArgs;

    UINT uMessage = GSM_CFH_GETCURRENTFILENAME; //MessageId

    //Input params

    InputArgs.CreateFromArgs2(
        CPackedArgs::ARGT_UINT32, &uMessage,
        CPackedArgs::ARGT_UINT32, &Handle
        );

    //-----------------------

    PVOID pInputBuffer = NULL;
    ULONG InputBufferSize = 0;

    if ( !InputArgs.Pack( &pInputBuffer, &InputBufferSize ) )
    {
        return ERROR_OUTOFMEMORY;
    }

    PVOID pOutputBuffer = NULL;
    ULONG OutputBufferSize = 0 ;

    UINT uReturnCode = CallGuestServer(
        ClientId,
        pInputBuffer,
        InputBufferSize,
        &pOutputBuffer,
        &OutputBufferSize
        );

    DbgPrintf( "HostCall_CFH_GetCurrentFileName: CallGuestServer return %u\n", uReturnCode );

    CPackedArgs::FreeMem( pInputBuffer );

    if ( uReturnCode != ERROR_SUCCESS )
        return uReturnCode;

    CPackedArgs OutputArgs;

    if ( !OutputArgs.Unpack( pOutputBuffer, OutputBufferSize ) )
    {
        DbgPrintf( "HostCall_CFH_GetCurrentFileName: Cannot unpack output buffer\n" );
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    //u32ErrorCode, PCOPY_FILES_GET_CURRENT_FILENAME

    if ( GSM_CFH_GETCURRENTFILENAME_OUT_CPARAMS != OutputArgs.GetCount() )
    {
        DbgPrintf( "HostCall_CFH_GetCurrentFileName: Bad output params count: %u\n", OutputArgs.GetCount() );
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    UINT u32ErrorCode = 0;
    PVOID pPtr = NULL;
    ULONG PtrSize = 0;

    if ( !OutputArgs.GetUInt32Arg( 0, &u32ErrorCode ) ||
         !OutputArgs.GetPtrArg( 1, &pPtr, &PtrSize ) ||
         PtrSize != sizeof(COPY_FILES_GET_CURRENT_FILENAME)
        )
    {
        DbgPrintf( "HostCall_CFH_GetCurrentFileName: Bad output params types\n", OutputArgs.GetCount() );

        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }


    RtlCopyMemory( pCopyFilesGetCurrentFileName, pPtr, sizeof(COPY_FILES_GET_CURRENT_FILENAME) );

    //Insure null terminator
    pCopyFilesGetCurrentFileName->wszRelFilePath[ ARRAYSIZE(pCopyFilesGetCurrentFileName->wszRelFilePath) - 1 ] = UNICODE_NULL;

    GS_FreeMem( pOutputBuffer );
    return u32ErrorCode;

}

UINT
HostCall_CFH_GetCurrentStream(
    IN  UINT ClientId,
	IN  UINT Handle,
	IN OUT PVOID pBuffer,
	IN  UINT BufferSize,
	OUT PUINT pBytesCopied,
	OUT PUINT64 pStreamOffset
	)
{
    CPackedArgs InputArgs;

    UINT uMessage = GSM_CFH_GETCURRENTSTREAM; //MessageId

    //u32Message, u32Handle u32BufferSize 

    //Input params

    InputArgs.CreateFromArgs3(
        CPackedArgs::ARGT_UINT32, &uMessage,
        CPackedArgs::ARGT_UINT32, &Handle,
        CPackedArgs::ARGT_UINT32, &BufferSize
        );

    //-----------------------

    PVOID pInputBuffer = NULL;
    ULONG InputBufferSize = 0;

    if ( !InputArgs.Pack( &pInputBuffer, &InputBufferSize ) )
    {
        return ERROR_OUTOFMEMORY;
    }

    PVOID pOutputBuffer = NULL;
    ULONG OutputBufferSize = 0 ;

    UINT uReturnCode = CallGuestServer(
        ClientId,
        pInputBuffer,
        InputBufferSize,
        &pOutputBuffer,
        &OutputBufferSize
        );

    CPackedArgs::FreeMem( pInputBuffer );

    if ( uReturnCode != ERROR_SUCCESS )
        return uReturnCode;

    CPackedArgs OutputArgs;

    if ( !OutputArgs.Unpack( pOutputBuffer, OutputBufferSize ) )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    //u32ErrorCode, ptrBuffer, u64Offset

    if ( GSM_CFH_GETCURRENTSTREAM_OUT_CPARAMS != OutputArgs.GetCount() )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    UINT u32ErrorCode = 0;
    PVOID pPtr = NULL;
    ULONG PtrSize = 0;
    UINT64 u64Offset = 0;

    if ( !OutputArgs.GetUInt32Arg( 0, &u32ErrorCode ) ||
         !OutputArgs.GetPtrArg( 1, &pPtr, &PtrSize ) ||
         PtrSize > BufferSize ||
         !OutputArgs.GetUInt64Arg( 2, &u64Offset )
        )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    RtlCopyMemory( pBuffer, pPtr, PtrSize );

    *pBytesCopied = PtrSize;
    *pStreamOffset = u64Offset;

    GS_FreeMem( pOutputBuffer );
    return u32ErrorCode;

}

UINT
HostCall_CFH_MoveToNextItem(
    IN  UINT ClientId,
	IN  UINT Handle,
	UINT64 ItemId
	)
{
    CPackedArgs InputArgs;

    UINT uMessage = GSM_CFH_MOVETONEXTITEM; //MessageId

    //u32Message, u32Handle, u64ItemId

    InputArgs.CreateFromArgs3(
        CPackedArgs::ARGT_UINT32, &uMessage,
        CPackedArgs::ARGT_UINT32, &Handle,
        CPackedArgs::ARGT_UINT64, &ItemId
        );

    //-----------------------

    PVOID pInputBuffer = NULL;
    ULONG InputBufferSize = 0;

    if ( !InputArgs.Pack( &pInputBuffer, &InputBufferSize ) )
    {
        return ERROR_OUTOFMEMORY;
    }

    PVOID pOutputBuffer = NULL;
    ULONG OutputBufferSize = 0 ;

    UINT uReturnCode = CallGuestServer(
        ClientId,
        pInputBuffer,
        InputBufferSize,
        &pOutputBuffer,
        &OutputBufferSize
        );

    CPackedArgs::FreeMem( pInputBuffer );

    if ( uReturnCode != ERROR_SUCCESS )
        return uReturnCode;

    CPackedArgs OutputArgs;

    if ( !OutputArgs.Unpack( pOutputBuffer, OutputBufferSize ) )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    //u32ErrorCode

    if ( GSM_CFH_MOVETONEXTITEM_OUT_CPARAMS != OutputArgs.GetCount() )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }


    UINT u32ErrorCode = 0;

    if ( !OutputArgs.GetUInt32Arg( 0, &u32ErrorCode ) )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    GS_FreeMem( pOutputBuffer );
    return u32ErrorCode;
}


UINT
HostCall_CFH_End(
	IN  UINT ClientId,
	IN  UINT Handle
	)
{
    CPackedArgs InputArgs;

    UINT uMessage = GSM_CFH_END; //MessageId

    //u32Message, u32Handle

    InputArgs.CreateFromArgs2(
        CPackedArgs::ARGT_UINT32, &uMessage,
        CPackedArgs::ARGT_UINT32, &Handle
        );

    //------------------------

    PVOID pInputBuffer = NULL;
    ULONG InputBufferSize = 0;

    if ( !InputArgs.Pack( &pInputBuffer, &InputBufferSize ) )
    {
        return ERROR_OUTOFMEMORY;
    }

    PVOID pOutputBuffer = NULL;
    ULONG OutputBufferSize = 0 ;

    UINT uReturnCode = CallGuestServer(
        ClientId,
        pInputBuffer,
        InputBufferSize,
        &pOutputBuffer,
        &OutputBufferSize
        );

    CPackedArgs::FreeMem( pInputBuffer );

    if ( uReturnCode != ERROR_SUCCESS )
        return uReturnCode;

    CPackedArgs OutputArgs;

    if ( !OutputArgs.Unpack( pOutputBuffer, OutputBufferSize ) )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    //u32ErrorCode

    if ( GSM_CFH_END_OUT_CPARAMS != OutputArgs.GetCount() )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    UINT u32ErrorCode = 0;

    if ( !OutputArgs.GetUInt32Arg( 0, &u32ErrorCode ) )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    GS_FreeMem( pOutputBuffer );
    return u32ErrorCode;

}

UINT
HostCall_DragDropFiles(
    UINT ClientId,
    UINT DropHandle
    )
{
    CPackedArgs InputArgs;

    UINT uMessage = GSM_DROP_FILES; //MessageId

    //u32Message, u32DropHandle

    InputArgs.CreateFromArgs2(
        CPackedArgs::ARGT_UINT32, &uMessage,
        CPackedArgs::ARGT_UINT32, &DropHandle
        );

    PVOID pInputBuffer = NULL;
    ULONG InputBufferSize = 0;

    if ( !InputArgs.Pack( &pInputBuffer, &InputBufferSize ) )
    {
        return ERROR_OUTOFMEMORY;
    }

    PVOID pOutputBuffer = NULL;
    ULONG OutputBufferSize = 0 ;

    UINT uReturnCode = CallGuestServer(
        ClientId,
        pInputBuffer,
        InputBufferSize,
        &pOutputBuffer,
        &OutputBufferSize
        );

    CPackedArgs::FreeMem( pInputBuffer );

    if ( uReturnCode != ERROR_SUCCESS )
        return uReturnCode;

    CPackedArgs OutputArgs;

    if ( !OutputArgs.Unpack( pOutputBuffer, OutputBufferSize ) )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    //u32ErrorCode

    if ( GSM_DROP_FILES_OUT_CPARAMS != OutputArgs.GetCount() )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    UINT u32ErrorCode = 0;

    if ( !OutputArgs.GetUInt32Arg( 0, &u32ErrorCode ) )
    {
        GS_FreeMem( pOutputBuffer );
        return ERROR_INVALID_PARAMETER;
    }

    GS_FreeMem( pOutputBuffer );
    return u32ErrorCode;
}

