/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

    
#include "vbox_hdrs.h"
#include <memory> 
#include <string>
#include <list> 
#include "NewClipService.h"
    
#include <windows.h>
#include "DbgPrintf.h"
#include "CopyFilesHandler.h"
#include "ParamsUtil.h"
#include "SecureFuncs.h"
#include "GServClient.h"
#include "DllMain.h"
#include "Globals.h"
#include "VMDragDrop.h"
#include "WinTools.h"
#include "VBoxUtil.h"
    
#define LogPrint(msg)   DbgPrintf msg //OutputDebugStringA msg
//VBoxGuestPropSvc_Orig.dll VBoxSharedClipboard_Orig.dll VBoxGuestControlSvc_Orig.dll
#define REDIR_LIB_NAMEA "VBoxGuestPropSvc_Orig.dll"

VOID
FilePathListToFileNamesList(
    vector<wstring> *pFilePathList,
    vector<wstring> *pFileNamesList
    );
    

static
DECLCALLBACK(void) 
svcCall(
    void *,
    VBOXHGCMCALLHANDLE callHandle,
    uint32_t u32ClientID,
    void *pvClient,
    uint32_t u32Function,
    uint32_t cParms,
    VBOXHGCMSVCPARM paParms[]
    );

static
DECLCALLBACK(int)
svcDisconnect(
    void *, 
    uint32_t u32ClientID, 
    void *pvClient
    );


static
DECLCALLBACK(int)
svcSaveState(
    void *pvService, 
    uint32_t u32ClientID, 
    void *pvClient, 
    PSSMHANDLE pSSM
    );
	
static 
DECLCALLBACK(int)
svcLoadState(
	void *pvService, 
	uint32_t u32ClientID, 
	void *pvClient, 
	PSSMHANDLE pSSM
	);
	
static
DECLCALLBACK(int)
svcConnect(
	void *pvService,
	uint32_t u32ClientID,
	void *pvClient
	);

typedef DECLCALLBACK(void)
TfnSvcCall(
    void *,
    VBOXHGCMCALLHANDLE callHandle,
    uint32_t u32ClientID,
    void *pvClient,
    uint32_t u32Function,
    uint32_t cParms,
    VBOXHGCMSVCPARM paParms[]
    );

typedef
DECLCALLBACK(int)
TfnSvcDisconnect(
    void *, 
    uint32_t u32ClientID, 
    void *pvClient
    );

typedef
DECLCALLBACK(int)
TfnSaveState(
    void *pvService, 
    uint32_t u32ClientID, 
    void *pvClient, 
    PSSMHANDLE pSSM
    );

typedef	
DECLCALLBACK(int)
TfnsvcLoadState(
	void *pvService, 
	uint32_t u32ClientID,
	void *pvClient,
	PSSMHANDLE pSSM
	);

typedef
DECLCALLBACK(int)
TfnsvcConnect(
	void *pvService,
	uint32_t u32ClientID,
	void *pvClient
	);


typedef TfnSvcCall* TpfnSvcCall;
typedef TfnSvcDisconnect* TpfnSvcDisconnect;
typedef TfnSaveState* TpfnSaveState;
typedef TfnsvcLoadState *TpfnsvcLoadState;
typedef TfnsvcConnect *TpfnsvcConnect;


HINSTANCE g_DllHinst = NULL;
HMODULE g_RedirLibHandle = NULL;
PFNVBOXHGCMSVCLOAD pfnVBoxHGCMSvcLoad = NULL;
TpfnSvcCall g_OrigSvcCall = NULL;
TpfnSvcDisconnect g_OrigSvcDisconnect = NULL;
TpfnSaveState g_OrigSaveState = NULL;
TpfnsvcLoadState g_OrigLoadState = NULL;
TpfnsvcConnect g_OrigSvcConnect = NULL;

PVBOXHGCMSVCHELPERS g_pHelpers = NULL;
wstring g_MachineName = L"";


DECLEXPORT(DWORD) HookVer = CURRENT_APP_VERSION;

EXTERN_C
BOOL 
WINAPI 
DllMain(
  HINSTANCE hinstDLL,
  DWORD fdwReason,
  LPVOID lpvReserved
  )
{
    if ( fdwReason == DLL_PROCESS_ATTACH )
    {
        
        DbgPrintf( "Attach library\n" );
        g_DllHinst = hinstDLL;
                
        g_RedirLibHandle = LoadLibraryA( REDIR_LIB_NAMEA );
        
        if ( NULL == g_RedirLibHandle )
        {
            LogPrint(( "Cannot load redir library\n" ));
            return FALSE;
        }
        
        pfnVBoxHGCMSvcLoad = (PFNVBOXHGCMSVCLOAD)GetProcAddress( g_RedirLibHandle, VBOX_HGCM_SVCLOAD_NAME );
        
        if ( NULL == pfnVBoxHGCMSvcLoad )
        {
            LogPrint(( "Cannot get address of VBOX_HGCM_SVCLOAD_NAME\n" ));
            return FALSE;
        }
    }
    else if ( fdwReason == DLL_PROCESS_DETACH )
    {
        VMDragDropClear();
		VBoxUtil_Clear();		
		

        DbgPrintf( "Free library\n" );
        
        if ( g_RedirLibHandle )
        {
            FreeLibrary( g_RedirLibHandle );
            g_RedirLibHandle = NULL;
        }
    }
    
    return TRUE;
}

extern "C" DECLCALLBACK(DECLEXPORT(int)) VBoxHGCMSvcLoad (VBOXHGCMSVCFNTABLE *ptable)
{
    int rc = VINF_SUCCESS;
    
    if ( ptable->cbSize != sizeof (VBOXHGCMSVCFNTABLE) ||
         ptable->u32Version != VBOX_HGCM_SVC_VERSION
       )
    {
        LogPrint(( "VBoxHGCMSvcLoad: Bad version or size\n" ));
        
        rc = pfnVBoxHGCMSvcLoad( ptable );
        
        return rc;
        //return VERR_VERSION_MISMATCH;
    }
    
    if ( pfnVBoxHGCMSvcLoad == NULL )
    {
        LogPrint(( "pfnVBoxHGCMSvcLoad is NULL\n" ));
        return VERR_VERSION_MISMATCH;
    }

    rc = pfnVBoxHGCMSvcLoad( ptable );
    
    LogPrint(( "VBoxHGCMSvcLoad started\n" ));
    
    if ( VINF_SUCCESS == rc )
    {
        LogPrint(( "Redirection stub was successfully called (ptable->pHelpers=%p)\n", ptable->pHelpers ));

        g_pHelpers = ptable->pHelpers;
        
        g_OrigSvcCall = ptable->pfnCall;
        ptable->pfnCall = svcCall;
        
        g_OrigSvcDisconnect = ptable->pfnDisconnect;
        ptable->pfnDisconnect = svcDisconnect;
        

        g_OrigSaveState = ptable->pfnSaveState;
        ptable->pfnSaveState = svcSaveState;
		
		g_OrigLoadState = ptable->pfnLoadState;
		ptable->pfnLoadState = svcLoadState;
		
		g_OrigSvcConnect = ptable->pfnConnect;
		ptable->pfnConnect    = svcConnect;
		
        CreateContextMenuCmdWorkerThread();

		VBoxUtil_Init();		
        VMDragDropInit();
		
		VBoxUtil_GetMachineName( &g_MachineName );
		
		DbgPrintf( "MachineName: %ws\n", g_MachineName.c_str() );

 
  }    
    
    
    return rc;
}

static
DECLCALLBACK(int)
svcDisconnect(
    void *ptr, 
    uint32_t u32ClientID, 
    void *pvClient
    )
{

    DbgPrintf( "Client %u was disconnected\n", u32ClientID );
    CFH_ClientDisconnected( u32ClientID );
    GSERVCLIENT_Handle_svcDisconnect( ptr, u32ClientID, pvClient );
    DbgPrintf( "Calling original stub: %p\n", g_OrigSvcDisconnect );
    
    int result = g_OrigSvcDisconnect( ptr, u32ClientID, pvClient );
    
    return result;
} 

static
DECLCALLBACK(void) 
svcCall(
    void *ptrUnk,
    VBOXHGCMCALLHANDLE callHandle,
    uint32_t u32ClientID,
    void *pvClient,
    uint32_t u32Function,
    uint32_t cParms,
    VBOXHGCMSVCPARM paParms[]
    )
{
    int rc = VINF_SUCCESS;
    
    
    //DbgPrintf( "svcCall: u32ClientID=%u u32Function=%u cParms=%u\n", u32ClientID, u32Function, cParms );
    
    if ( g_OrigSvcCall == NULL || g_pHelpers == NULL )
    {
        if ( g_pHelpers )
        {
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
        }
        
        return;
    }

    if ( GSERVCLIENT_Handle_svcCall( g_pHelpers, ptrUnk,callHandle,u32ClientID,pvClient,u32Function,cParms, paParms ) )
        return;
    
    if ( VBOX_SHARED_CFH_BEGIN == u32Function )
    {
        LogPrint(( "VBOX_SHARED_CFH_BEGIN called\n" ));
        
        if ( VBOX_SHARED_CFH_BEGIN_CPARAMS != cParms )
        {
            LogPrint(( "VBOX_SHARED_CFH_BEGIN: Invalid parameters count\n" ));
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }
		
		UINT uClipBoardMode = CLIPBOARD_MODE_BIDIRECTIONAL;
		
		if ( VBoxUtil_GetClipboardMode(  &uClipBoardMode ) &&
			 !( uClipBoardMode == CLIPBOARD_MODE_BIDIRECTIONAL || uClipBoardMode == CLIPBOARD_MODE_HOST_TO_GUEST ) 
			)
		{
			g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_STATE );
			return;
		}
        
        //uErrorCode, uDropHandle, uHandle

        UINT uDropHandle = 0;
        
        if (  paParms[0].type != VBOX_HGCM_SVC_PARM_32BIT ||
              !VBoxUINT32ParamGet( &paParms[1], &uDropHandle ) ||
              paParms[2].type != VBOX_HGCM_SVC_PARM_32BIT
            )
        {
            LogPrint(( "VBOX_SHARED_CFH_BEGIN: Invalid parameters type\n" ));
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }
        
        uint32_t ErrorCode = 0;
        uint32_t ResultHandle = 0;
        
        ErrorCode = CFH_Begin( u32ClientID, uDropHandle, &ResultHandle );
        
        VBoxUINT32ParamSet( &paParms[0], ErrorCode );
        VBoxUINT32ParamSet( &paParms[2], ResultHandle );
                
        g_pHelpers->pfnCallComplete ( callHandle, rc );
        return;
    }
    else if ( VBOX_SHARED_CFH_GET_BUILDING_LIST_STATUS == u32Function )
    {
        if ( VBOX_SHARED_CFH_GET_BUILDING_LIST_STATUS_CPARAMS != cParms )
        {
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }
        
        //uErrorCode, uHandle, PCOPY_FILES_GET_CURRENT_FILENAME
        uint32_t ErrorCode = 0;        
        uint32_t InputHandle = 0;
        PCOPY_FILES_BUILDING_LIST_STATUS pStatus = NULL;
        uint32_t StructSize = 0;
        
        if (
            paParms[0].type != VBOX_HGCM_SVC_PARM_32BIT ||
            !VBoxUINT32ParamGet( &paParms[1], &InputHandle ) ||
            !VBoxPointerParamGet( &paParms[2], (void **)&pStatus, &StructSize  ) ||
            StructSize != sizeof(COPY_FILES_BUILDING_LIST_STATUS)
            )
        {
            LogPrint(( "VBOX_SHARED_CFH_GET_BUILDING_LIST_STATUS: Invalid parameters type\n" ));
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }
        
        ErrorCode = CFH_GetBuildingListStatus(
            u32ClientID,
            InputHandle,
            pStatus
            );
        
        VBoxUINT32ParamSet( &paParms[0], ErrorCode );
        g_pHelpers->pfnCallComplete ( callHandle, rc );
        return;
    }
    else if ( VBOX_SHARED_CFH_GETCURRENTFILENAME == u32Function )
    {
        uint32_t ErrorCode = 0;
        uint32_t InputHandle = 0;
        PCOPY_FILES_GET_CURRENT_FILENAME pGetCurrentFileName = NULL;
        uint32_t StructSize = 0;
        
        DbgPrintf( "VBOX_SHARED_CFH_GETCURRENTFILENAME\n" );
        
        if ( VBOX_SHARED_CFH_GETCURRENTFILENAME_CPARAMS != cParms )
        {
            DbgPrintf( "Bad count of parameters\n" );
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }
        
        //ErrorCode, Handle, PCOPY_FILES_GET_CURRENT_FILENAME
        
        if (
            paParms[0].type != VBOX_HGCM_SVC_PARM_32BIT ||
            !VBoxUINT32ParamGet( &paParms[1], &InputHandle ) ||
            !VBoxPointerParamGet( &paParms[2], (void **)&pGetCurrentFileName, &StructSize  ) ||
            StructSize != sizeof(COPY_FILES_GET_CURRENT_FILENAME)
            )
        {
            DbgPrintf( "VBOX_SHARED_CFH_GETCURRENTFILENAME: Invalid parameters type\n" );
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }
        
        pGetCurrentFileName->wszRelFilePath[ ARRAYSIZE(pGetCurrentFileName->wszRelFilePath) - 1 ] = 
            UNICODE_NULL;
        
        ErrorCode = CFH_GetCurrentFileName(
            u32ClientID,
            InputHandle,
            pGetCurrentFileName
            );
        
        DbgPrintf( "CFH_GetCurrentFileName return %u\n", ErrorCode );
        
        if ( ERROR_SUCCESS == ErrorCode )
        {
            DbgPrintf( "Current File: %ws\n", pGetCurrentFileName->wszRelFilePath );
        }
        
        VBoxUINT32ParamSet( &paParms[0], ErrorCode );
        g_pHelpers->pfnCallComplete ( callHandle, rc );
        return;
    }
    else if ( VBOX_SHARED_CFH_GET_CURRENTSTREAM == u32Function )
    {
        
        if ( VBOX_SHARED_CFH_GET_CURRENTSTREAM_CPARAMS != cParms )
        {
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }
        
        //OUT ErrorCode, IN InputHandle, IN OUT <Buffer&Size>, OUT uCopied, OUT u64StreamOffset
        uint32_t ErrorCode = 0;
        uint32_t InputHandle = 0;
        PVOID pBuffer = NULL;
        uint32_t BufferSize = 0;
        uint32_t uCopied = 0;
        uint64_t u64StreamOffset = 0;
        
        if (
            paParms[0].type != VBOX_HGCM_SVC_PARM_32BIT ||
            !VBoxUINT32ParamGet( &paParms[1], &InputHandle ) ||
            !VBoxPointerParamGet( &paParms[2], &pBuffer, &BufferSize ) ||
            BufferSize == 0 ||
            paParms[3].type != VBOX_HGCM_SVC_PARM_32BIT ||
            paParms[4].type != VBOX_HGCM_SVC_PARM_64BIT
            )
        {
            LogPrint(( "VBOX_SHARED_CFH_GET_CURRENTSTREAM: Invalid parameters type\n" ));            
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }
        
        ErrorCode = CFH_GetCurrentStream(
            u32ClientID,
            InputHandle,
            pBuffer,
            BufferSize,
            &uCopied,
            &u64StreamOffset
            );
        
        VBoxUINT32ParamSet( &paParms[3], uCopied );
        VBoxUINT64ParamSet( &paParms[4], u64StreamOffset );
        
        VBoxUINT32ParamSet( &paParms[0], ErrorCode );
        g_pHelpers->pfnCallComplete ( callHandle, rc );
        return;
    }
    else if (  VBOX_SHARED_MOVE_TO_NEXT_ITEM == u32Function )
    {
        if ( VBOX_SHARED_MOVE_TO_NEXT_ITEM_CPARAMS != cParms )
        {
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }

        //ErrorCode, Handle, NextItemId

        uint32_t ErrorCode = 0;
        uint32_t InputHandle = 0;
        uint64_t NextItemId = 0;

        if ( 
            paParms[0].type != VBOX_HGCM_SVC_PARM_32BIT ||
            !VBoxUINT32ParamGet( &paParms[1], &InputHandle ) ||
            !VBoxUINT64ParamGet( &paParms[2], &NextItemId )
            )
        {
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }

        ErrorCode = CFH_MoveToNextItem(
            u32ClientID,
            InputHandle,
            NextItemId
            );

        VBoxUINT32ParamSet( &paParms[0], ErrorCode );
        
        g_pHelpers->pfnCallComplete ( callHandle, rc );
        return;


    }
    else if ( VBOX_SHARED_CFH_END == u32Function )
    {
        LogPrint(( "VBOX_SHARED_CFH_END called\n" ));
        
        if ( VBOX_SHARED_CFH_END_CPARAMS != cParms )
        {
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }
        
        
        //ErrorCode, Handle
        
        uint32_t ErrorCode = 0;
        uint32_t InputHandle = 0;
                
        if ( 
            paParms[0].type != VBOX_HGCM_SVC_PARM_32BIT ||
            !VBoxUINT32ParamGet( &paParms[1], &InputHandle ) 
            )
        {
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }

        ErrorCode = CFH_End(
            u32ClientID,
            InputHandle
            );
            
        VBoxUINT32ParamSet( &paParms[0], ErrorCode );
        
        g_pHelpers->pfnCallComplete ( callHandle, rc );
        return;
    }
    else if ( VBOX_SHARED_GET_VERSION == u32Function )
    {
        if ( VBOX_SHARED_GET_VERSION_CPARAMS != cParms ||
             paParms[0].type != VBOX_HGCM_SVC_PARM_32BIT
            )
        {
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }

        VBoxUINT32ParamSet( &paParms[0], CURRENT_APP_VERSION );
        g_pHelpers->pfnCallComplete ( callHandle, rc );
        return;
    }
    else if ( VBOX_SHARED_GET_DROPPED_FILES_LIST == u32Function )
    {
        UINT u32DropHandle = 0;
        PVOID pFileListBuff = NULL;
        uint32_t FileListBuffSize = 0;
        UINT u32BuffWriteln = 0;

        //IN u32DropHandle, out pFileListBuff, out u32FileListBuffWriteln

        if ( VBOX_SHARED_GET_DROPPED_FILES_LIST_CPARAMS != cParms ||
             !VBoxUINT32ParamGet( &paParms[0], &u32DropHandle ) ||
             !VBoxPointerParamGet( &paParms[1], (void **)&pFileListBuff, &FileListBuffSize  ) ||
             FileListBuffSize == 0 ||
             paParms[2].type != VBOX_HGCM_SVC_PARM_32BIT
            )
        {
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }
		
		UINT uClipBoardMode = CLIPBOARD_MODE_BIDIRECTIONAL;
		
		if ( VBoxUtil_GetClipboardMode(  &uClipBoardMode ) &&
			 !( uClipBoardMode == CLIPBOARD_MODE_BIDIRECTIONAL || uClipBoardMode == CLIPBOARD_MODE_HOST_TO_GUEST ) 
			)
		{
			g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_STATE );
			return;
		}		

        vector<wstring> _DroppedFileList;

        if ( !GetDroppedFilesListByHandle( u32DropHandle, &_DroppedFileList ) )
        {
            DbgPrintf( "GetDroppedFilesListByHandle failed\n" );
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }

        vector<wstring> FileNamesList;

        FilePathListToFileNamesList(
            &_DroppedFileList,
            &FileNamesList
            );

        ULONG RequiredBuffSize = 0;

        if ( !StringListToBuffW( &FileNamesList, pFileListBuff, (ULONG)FileListBuffSize, &RequiredBuffSize, (PULONG)&u32BuffWriteln  ) )
        {
            DbgPrintf( "StringListToBuffW failed\n" );
            g_pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
            return;
        }

        VBoxUINT32ParamSet( &paParms[2], u32BuffWriteln );
        g_pHelpers->pfnCallComplete ( callHandle, rc );
        return;
    }

    
    g_OrigSvcCall(
        ptrUnk,
        callHandle,
        u32ClientID,
        pvClient,
        u32Function,
        cParms,
        paParms
        );
    
}

HINSTANCE 
CALLBACK 
GetModuleInstance()
{
    return g_DllHinst;
}


DECLCALLBACK(int)
svcSaveState(
    void *pvService, 
    uint32_t u32ClientID, 
    void *pvClient, 
    PSSMHANDLE pSSM
    )
{
    GSERVCLIENT_Handle_svcSaveState( pvService, u32ClientID, pvClient, pSSM );

    if ( g_OrigSaveState != NULL )
    {
        return g_OrigSaveState( pvService, u32ClientID, pvClient, pSSM );
    }

    return VINF_SUCCESS;
}


VOID
FilePathListToFileNamesList(
    vector<wstring> *pFilePathList,
    vector<wstring> *pFileNamesList
    )
{
    pFileNamesList->clear();

    for (
        vector<wstring>::iterator iter = pFilePathList->begin();
        iter != pFilePathList->end();
        iter++
        )
    {
        LPCWSTR psz = wcsrchr( iter->c_str(), L'\\' );

        if ( psz )
            pFileNamesList->push_back( psz + 1 );
        else
            pFileNamesList->push_back( iter->c_str() );

    }
}

static
DECLCALLBACK(int)
svcLoadState(
	void *pvService, 
	uint32_t u32ClientID, 
	void *pvClient, 
	PSSMHANDLE pSSM
	)
{
	
	GSERVCLIENT_Handle_svcLoadState( pvService, u32ClientID, pvClient, pSSM );
	
	if ( g_OrigLoadState != NULL )
	{
		return g_OrigLoadState( pvService, u32ClientID, pvClient, pSSM );
	}

	return VINF_SUCCESS;
}

static
DECLCALLBACK(int)
svcConnect(
	void *pvService,
	uint32_t u32ClientID,
	void *pvClient
	)
{
	GSERVCLIENT_Handle_svcConnect( pvService, u32ClientID, pvClient );
	
	if ( g_OrigSvcConnect != NULL )
	{
		return g_OrigSvcConnect( pvService, u32ClientID, pvClient );
	}

	return VINF_SUCCESS;
}

