/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/




#include "DragDropGuest.h"
#include "PathFromPoint.h"
#include "DbgPrintf.h"
#include "WinTools.h"
#include "CriticalSection.h"
#include "resource.h"
#include "DllMain.h"
#include "GuestCalls.h"
#include "GuestServiceConnect.h"
#include "LangTools.h"
#include "DropFilesDataObject.h"
#include <shlobj.h> //DROPFILES
#include <windowsX.h>
#include "Globals.h"
#include <strsafe.h>
#include "Settings.h"

//---------

typedef struct _COPY_DROPPED_FILES_DLGPARAMS
{
    wstring DefaultPath;
    wstring ResultPath;

} COPY_DROPPED_FILES_DLGPARAMS, *PCOPY_DROPPED_FILES_DLGPARAMS;

//---------


VOID
DragDrop_CopyFilesToFolder(
    UINT uDropHandle,
    LPCWSTR lpszFolder
    );

VOID
Ole_SimulateDragDrop(
    UINT uDropHandle,
    HWND hWndFromPoint,
    POINT pt
    );


VOID
DragDrop_Simulate_WM_DROPFILES(
    UINT uDropHandle,
    HWND hWndFromPoint,
    POINT pt
    );

VOID
DragDrop_ShowSelectFolderDlg(
    UINT uDropHandle,
    HWND hWndFromPoint,
    POINT pt
    );


INT_PTR
CALLBACK
CopyDroppedFilesDialogProc(
    HWND hwndDlg,
    UINT uMsg,
    WPARAM wParam,
    LPARAM lParam
    );


VOID
GSM_AsyncOnDropFiles( UINT uDropHandle );

DWORD
WINAPI
OnDropFilesThread(
    LPVOID lpParameter
    );

VOID
OnDropFilesFromHostInternal( UINT uDropHandle );

//---------

CCriticalSection g_DragDropLock;

//---------


VOID
GSM_AsyncOnDropFiles( UINT uDropHandle )
{
    DWORD dwTid = 0;

    HANDLE hThread = CreateThread(
        NULL,
        0,
        (LPTHREAD_START_ROUTINE)OnDropFilesThread,
        (PVOID)uDropHandle, //parameter
        0,
        &dwTid
        );

    CloseHandle( hThread );
    hThread = NULL;

}

DWORD
WINAPI
OnDropFilesThread(
    LPVOID lpParameter
    )
{
    UINT uDropHandle = (UINT)lpParameter;

    if ( !g_DragDropLock.TryLock() )
    {
        DbgPrintf( "Other DRAGDROP event processing\n" );
        return -1;
    }

    OnDropFilesFromHostInternal( uDropHandle );
    g_DragDropLock.Unlock();


    return 0;

}

VOID
OnDropFilesFromHostInternal( UINT uDropHandle )
{
    DbgPrintf( "OnDropFilesFromHostInternal(%u)\n", uDropHandle );


    POINT pt = {0,0};

    GetCursorPos( &pt );

    HWND hWndFromPoint = WindowFromPoint( pt );
    DWORD dwWndFromPointPID = 0;

    GetWindowThreadProcessId( hWndFromPoint, &dwWndFromPointPID );

    
    DbgPrintf(
        "Drop pt.x=%u pt.y=%u hwnd=%p pid=%u(0x%x)\n", 
        (ULONG)pt.x, 
        (ULONG)pt.y, 
        hWndFromPoint, 
        dwWndFromPointPID, 
        dwWndFromPointPID 
        );


    /*DragDrop_ShowSelectFolderDlg(
        uDropHandle,
        hWndFromPoint,
        pt
        );*/

    Ole_SimulateDragDrop(
        uDropHandle,
        hWndFromPoint,
        pt
        );



}

INT_PTR
CALLBACK
CopyDroppedFilesDialogProc(
    HWND hwndDlg,
    UINT uMsg,
    WPARAM wParam,
    LPARAM lParam
    )
{
    switch ( uMsg )
    {
        case WM_INITDIALOG:
        {
            PCOPY_DROPPED_FILES_DLGPARAMS pParams = (PCOPY_DROPPED_FILES_DLGPARAMS)lParam;

            SetWindowLongPtr(
                hwndDlg,
                DWLP_USER,
                (LONG_PTR)pParams
                );

            ComboBox_AddString(
                GetDlgItem( hwndDlg, ICB_FOLDER ),
                UnicodeToAnsi( pParams->DefaultPath ).c_str()
                );

            ComboBox_SetCurSel(
                GetDlgItem( hwndDlg, ICB_FOLDER ),
                0
                );

            SetForegroundWindow( hwndDlg );

            return TRUE;
            break;
        }

        case WM_COMMAND:
        {
            switch ( LOWORD(wParam) )
            {
                case IDCANCEL:
                {
                    EndDialog( hwndDlg, FALSE );
                    break;
                }

                case IDOK:
                {
                    PCOPY_DROPPED_FILES_DLGPARAMS pParams = (PCOPY_DROPPED_FILES_DLGPARAMS)GetWindowLongPtr(
                        hwndDlg,
                        DWLP_USER
                        );

                    if ( NULL == pParams )
                        break;

                    int CurSel = ComboBox_GetCurSel( GetDlgItem( hwndDlg, ICB_FOLDER ) );
    
                    if ( CurSel == CB_ERR )
                        break;

                    WCHAR wszFolder[MAX_PATH] = L"";

                    GetWindowTextW( GetDlgItem( hwndDlg, ICB_FOLDER ), wszFolder, ARRAYSIZE(wszFolder) );

                    if ( wszFolder[0] == 0 )
                        break;

                    pParams->ResultPath = wszFolder;

                    EndDialog( hwndDlg, TRUE );
                    break;
                }
                
            }

            break;
        }
    }

    return FALSE;
}

VOID
DragDrop_CopyFilesToFolder(
    UINT uDropHandle,
    LPCWSTR lpszFolder
    )
{
    CGuestFromHostCopierHelper CopyingHelper;

    UINT ConnectionId = 0;

    int code = CONNECT_FUNCTION( (PUINT)&ConnectionId );

    if ( RT_SUCCESS( code ) )
    {
        ULONG uRet = ProceeedCopyingFiles( &CopyingHelper, ConnectionId, uDropHandle, lpszFolder );
		
		if ( uRet > 0 && Settings_BeepAfterCopyingFiles() )
		{
			MessageBeep( MB_OK );
		}
		
        DISCONNECT_FUNCTION( ConnectionId );
    }
}


VOID
DragDrop_ShowSelectFolderDlg(
    UINT uDropHandle,
    HWND hWndFromPoint,
    POINT pt
    )
{
    wstring wPath = PathFromPoint( pt );

    if ( wPath == L"" )
    {
        wPath = GetDesktopFolder();
    }

    COPY_DROPPED_FILES_DLGPARAMS DlgParams;

    DlgParams.DefaultPath = wPath;

    INT_PTR DlgResult = DialogBoxParam(
        g_DllHinst,
        MAKEINTRESOURCE( IDD_COPY_DROPPED_FILES ),
        NULL,
        CopyDroppedFilesDialogProc,
        (LPARAM)&DlgParams
        );

    if ( !DlgResult || DlgParams.ResultPath == L"" )
        return;

    DragDrop_CopyFilesToFolder( uDropHandle, DlgParams.ResultPath.c_str() );

}

VOID
DragDrop_Simulate_WM_DROPFILES(
    UINT uDropHandle,
    HWND hWndFromPoint,
    POINT pt
    )
{
    //CLSID_DragDropHelper IID_IDropTargetHelper (IDropSourceHelper) (RegisterDragDrop  OleInitialize)

    DWORD dwWindowStyleEx = GetWindowLongPtr( hWndFromPoint, GWL_EXSTYLE );
    HWND hwndTargetWindow = NULL;

    if ( (dwWindowStyleEx & WS_EX_ACCEPTFILES) == 0 )
    {
        DbgPrintf( "WARNING: WS_EX_ACCEPTFILES is not present in target window(%p)\n", hWndFromPoint );

        //Find parent window with WS_EX_ACCEPTFILES style?

        HWND hwndParent = hWndFromPoint;

        while ( (hwndParent = GetParent( hwndParent )) != NULL )
        {
            dwWindowStyleEx = GetWindowLongPtr( hwndParent, GWL_EXSTYLE );

            if ( (dwWindowStyleEx & WS_EX_ACCEPTFILES) != 0 )
            {
                hwndTargetWindow = hwndParent;
                break;
            }
        }


        if ( NULL == hwndTargetWindow )
        {
            LangErrorMessageBox( IDS_ERR_DRAGDROP_NOT_SUPPORTED );
            return;
        }

        DWORD dwTmpPid = 0;
        GetWindowThreadProcessId( hwndTargetWindow, &dwTmpPid );

        DbgPrintf( "Found parent window (%p) with WS_EX_ACCEPTFILES style pid=%u\n", hwndTargetWindow, dwTmpPid  );
    }
    else
    {
        hwndTargetWindow = hWndFromPoint;
    }

    wstring wPath = PathFromPoint( pt );

    if ( wPath != L"" )
    {
        DragDrop_CopyFilesToFolder( uDropHandle, wPath.c_str() );
        return;
    }

   //Copy files to temp folder

    WCHAR wszTempPath[ MAX_PATH ] = L""; 
    WCHAR wszTempFolderName[ MAX_PATH ] = L"";

    if ( 0 == GetTempPathW( ARRAYSIZE(wszTempPath), wszTempPath ) ||
         GetTempFileNameW( wszTempPath,
                              TEMP_FOLDER_PREFIXW,
                              0,
                              wszTempFolderName ) == 0
        )
    {
        DbgPrintf( "Cannot generate temp file name: %u\n", GetLastError() );
        LangErrorMessageBox( IDS_CANNOT_CREATE_TEMP_FOLDER );
        return;
    }

    //Delete previously created file

    DeleteFileW( wszTempFolderName );

    if ( !CreateDirectoryW( wszTempFolderName, NULL ) )
    {
        DbgPrintf( "Cannot CreateDirectory(%ws) : %u\n", wszTempFolderName, GetLastError() );
        LangErrorMessageBox( IDS_CANNOT_CREATE_TEMP_FOLDER );
        return;
    }

    DbgPrintf( "Temp folder (%ws) was created \n", wszTempFolderName );

    DragDrop_CopyFilesToFolder( uDropHandle, wszTempFolderName );

    //Get Temp Dir files list

    vector<wstring> FilesList;

    GetDirFilesListW( wszTempFolderName, &FilesList );

    if ( FilesList.size() == 0 )
    {
        DbgPrintf( "Temp folder (%ws) does not contain files\n", wszTempFolderName );
        return;
    }

    UINT BufferSize = sizeof(DROPFILES) + 2*sizeof(WCHAR);

    for (
        vector<wstring>::iterator iter = FilesList.begin();
        iter != FilesList.end();
        iter++
        )
    {
        BufferSize += ( (iter->length() + 1 ) * sizeof(WCHAR) );
    } 
    

    HGLOBAL hNewDropHandle = GlobalAlloc( GHND | GMEM_SHARE, BufferSize  );

    if ( NULL == hNewDropHandle )
    {
        DbgPrintf( "GlobalAlloc failed\n" );
        return;
    }

    LPDROPFILES pDropFiles = (LPDROPFILES)GlobalLock( hNewDropHandle );

    if ( NULL == pDropFiles )
    {
        DbgPrintf( "GlobalLock failed\n" );
        GlobalFree( hNewDropHandle );
        return;
    }

    pDropFiles->fWide = TRUE;
    pDropFiles->pt = pt;
    pDropFiles->fNC = TRUE;
    pDropFiles->pFiles = sizeof( DROPFILES );

    PUCHAR pszCurr = (PUCHAR) ( (PUCHAR)pDropFiles + pDropFiles->pFiles  );

    for (
        vector<wstring>::iterator iter = FilesList.begin();
        iter != FilesList.end();
        iter++
        )
    {
        RtlCopyMemory(
            pszCurr,
            (PVOID)iter->c_str(),
            (iter->length()+1)*sizeof(WCHAR)
            );

        pszCurr += (iter->length()+1)*sizeof(WCHAR);
    } 

    //Last null terminators

    pszCurr[0] = 0;
    pszCurr[1] = 0;

    pszCurr[2] = 0;
    pszCurr[3] = 0;


    GlobalUnlock( hNewDropHandle );

    //SendMessage does not work?

    PostMessage( hwndTargetWindow, WM_DROPFILES, (WPARAM)hNewDropHandle, 0 );

    //FIXME:
    Sleep(60);
    GlobalFree( hNewDropHandle );
}

wstring
GetDragDropTmpMainFolder()
{
	WCHAR wszTempPath[ MAX_PATH ] = L""; 
	
	if ( 0 == GetTempPathW( ARRAYSIZE(wszTempPath), wszTempPath ) )
		return L"";
		
	int len = wcslen( wszTempPath );
	
	if ( len > 0 && wszTempPath[ len-1 ] != L'\\' )		
		StringCchCatW( wszTempPath, ARRAYSIZE(wszTempPath), L"\\" );
		
	StringCchCatW( wszTempPath, ARRAYSIZE(wszTempPath), GUEST_APP_NAMEW );
	
	return wszTempPath;
	
}

wstring
CreateDragDropTmpFolder()
{
    WCHAR wszTempPath[ MAX_PATH ] = L""; 
    WCHAR wszTempFolderName[ MAX_PATH ] = L"";
	
	wstring wstr = GetDragDropTmpMainFolder();
	
	if ( wstr.length() == 0 )
	{
        DbgPrintf( "Cannot get temp folder name\n" );
        return L"";	
	}
	
	CreateDirectoryW( wstr.c_str(), NULL );
	StringCchCopyW(  wszTempPath, ARRAYSIZE(wszTempPath), wstr.c_str() );
	
    if ( GetTempFileNameW( wszTempPath,
                              TEMP_FOLDER_PREFIXW,
                              0,
                              wszTempFolderName ) == 0
        )
    {
        DbgPrintf( "Cannot generate temp file name: %u\n", GetLastError() );
        return L"";
    }

    //Delete previously created file by GetTempFileNameW function

    DeleteFileW( wszTempFolderName );

    if ( !CreateDirectoryW( wszTempFolderName, NULL ) )
    {
        DbgPrintf( "Cannot CreateDirectory(%ws) : %u\n", wszTempFolderName, GetLastError() );        
        return L"";
    }

    return wstring( wszTempFolderName );
}


VOID
Ole_SimulateDragDrop(
    UINT uDropHandle,
    HWND hWndFromPoint,
    POINT pt
    )
{
    vector<wstring> _FileNamesList;
    BOOL bRet = FALSE;

    DbgPrintf( "Ole_SimulateDragDrop\n" );

    GuestCall_GetDroppedFileList(
        0, //ClientId - autoconnect
        uDropHandle,
        &_FileNamesList
        );

    if ( _FileNamesList.size() == 0 )
        return;

    wstring wTmpDir = CreateDragDropTmpFolder();

    if ( wTmpDir == L"" )
    {
        LangErrorMessageBox( IDS_CANNOT_CREATE_TEMP_FOLDER );
        return;
    }

    OleInitialize( NULL );

    vector<wstring> FilePathList;

    for (
        vector<wstring>::iterator iter = _FileNamesList.begin();
        iter != _FileNamesList.end();
        iter++
        )
    {
        wstring wPath;

        wPath =  wTmpDir + wstring( L"\\" ) + *iter;

        DbgPrintf( "[%ws],%u=>%ws,%u\n", iter->c_str(), (ULONG)iter->length(), wPath.c_str(), wPath.length() );

        FilePathList.push_back( wPath );
    }

    CDropFilesDataObject *pDropFilesDataObject = new CDropFilesDataObject( 
        uDropHandle,
        &FilePathList,
        pt,
        wTmpDir.c_str()
        );

    if ( NULL == pDropFilesDataObject )
    {
        OleUninitialize();
        return;
    }

    CDropSource *pDropSource = new CDropSource( pDropFilesDataObject );

    if ( NULL == pDropSource )
    {
        pDropFilesDataObject->Release();
        OleUninitialize();
        return;
    }

    DWORD dwEffect = 0;

    //FIXME: TODO: improve:  it needs because DoDragDrop use PeekMessage loop for quering WM_MOUSE*** message

    DbgPrintf( "PUT WM_MOUSEMOVE into thread messages queue\n" );

    PostThreadMessage( GetCurrentThreadId(), WM_MOUSEMOVE, 0, MAKELONG( pt.x, pt.y ) );

    DbgPrintf( "Call DoDragDrop\n" );

    HRESULT hr = DoDragDrop(
      pDropFilesDataObject,
      pDropSource,
      DROPEFFECT_COPY | DROPEFFECT_LINK,
      &dwEffect
      );

    //DRAGDROP_S_DROP DRAGDROP_S_CANCEL
    DbgPrintf( "DoDragDrop: 0x%x\n", (ULONG)hr );

    pDropSource->Release();
    pDropFilesDataObject->Release();

    OleUninitialize();
}

VOID
TestDragDrop( POINT pt )
{

    BOOL bRet = FALSE;

    HRESULT hr = OleInitialize( NULL );

    DbgPrintf( "OleInitialize return %x\n", hr );

    vector<wstring> FilePathList;

    wstring wWinDir;
    WCHAR szWinDir[ MAX_PATH + 1 ] = L"";
    GetWindowsDirectoryW( szWinDir, ARRAYSIZE(szWinDir) );

    wWinDir = szWinDir;

    FilePathList.push_back( wWinDir + L"\\winhlp32.exe" );

    CDropFilesDataObject *pDropFilesDataObject = new CDropFilesDataObject( 
        0,
        &FilePathList,
        pt,
        szWinDir //wTmpDir.c_str()
        );

    if ( NULL == pDropFilesDataObject )
    {
        OleUninitialize();
        return;
    }

    
    DbgPrintf( "Call Test_SetDroppedFilesWasCopied\n" );

    pDropFilesDataObject->Test_SetDroppedFilesWasCopied( TRUE );

    CDropSource *pDropSource = new CDropSource( pDropFilesDataObject );

    if ( NULL == pDropSource )
    {
        pDropFilesDataObject->Release();
        OleUninitialize();
        return;
    }

    DWORD dwEffect = 0;

    DbgPrintf( "PUT WM_MOUSEMOVE into thread messages queue\n" );

    PostThreadMessage( GetCurrentThreadId(), WM_MOUSEMOVE, 0, MAKELONG( pt.x, pt.y ) );

    DbgPrintf( "Call DoDragDrop\n" );

    hr = DoDragDrop(
      pDropFilesDataObject,
      pDropSource,
      DROPEFFECT_COPY | DROPEFFECT_LINK,
      &dwEffect
      );

    //DRAGDROP_S_DROP DRAGDROP_S_CANCEL
    DbgPrintf( "DoDragDrop: 0x%x\n", (ULONG)hr );

    pDropSource->Release();
    pDropFilesDataObject->Release();

    OleUninitialize();



}


