/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "DlgWaitProgress.h"

#ifndef PROGDLG_MARQUEEPROGRESS
#define PROGDLG_MARQUEEPROGRESS 0x00000020
#endif


COM_DlgWait::COM_DlgWait()
{
    m_ProgressDlg = NULL;
    m_DlgCreateFailed = FALSE;
    m_DlgShowed = FALSE;
}

COM_DlgWait::~COM_DlgWait()
{
    if ( m_ProgressDlg )
    {
        m_ProgressDlg->StopProgressDialog();
        m_ProgressDlg->Release();
        m_ProgressDlg = NULL;
    }
}

BOOL
COM_DlgWait::IsCancelled()
{
    if ( m_ProgressDlg == NULL )
        return FALSE;

    return m_ProgressDlg->HasUserCancelled();
}

BOOL
COM_DlgWait::IsVistaOrLater()
{
    OSVERSIONINFOEX OsVersionInfoEx;

    RtlZeroMemory( &OsVersionInfoEx, sizeof(OsVersionInfoEx) );
    OsVersionInfoEx.dwOSVersionInfoSize = sizeof(OsVersionInfoEx);

    GetVersionEx( (OSVERSIONINFO *)&OsVersionInfoEx );

    if ( OsVersionInfoEx.dwMajorVersion >= 6 )
        return TRUE;
    else
        return FALSE;

}

VOID
COM_DlgWait::EndDialog()
{
    if ( m_ProgressDlg == NULL )
        return;

    if ( m_DlgShowed )
    {
        m_ProgressDlg->StopProgressDialog();
        m_DlgShowed = FALSE;
    }

}

BOOL
COM_DlgWait::Show(
    HWND hWndParent,
    LPCWSTR lpszTitle,
    LPCWSTR lpszLineMessage,
    LPCWSTR lpszCancelMessage
    )
{
    if ( m_DlgCreateFailed )
        return FALSE;

    if ( m_ProgressDlg == NULL )
    {
        HRESULT hr = CoCreateInstance(
            CLSID_ProgressDialog, 
            NULL, 
            CLSCTX_INPROC_SERVER, 
            IID_IProgressDialog, 
            (void **)&m_ProgressDlg
            );

        if ( FAILED(hr) )
        {
            m_DlgCreateFailed = TRUE;
            m_ProgressDlg = NULL;
            return FALSE;
        }

        m_ProgressDlg->Timer( PDTIMER_RESET, NULL );

        m_ProgressDlg->SetTitle( lpszTitle );
        m_ProgressDlg->SetLine( 1, lpszLineMessage, FALSE, NULL );
        m_ProgressDlg->SetCancelMsg( lpszCancelMessage, NULL );

        DWORD dwFlags = PROGDLG_NOTIME;

        if ( IsVistaOrLater() )
        {
            dwFlags |= PROGDLG_MARQUEEPROGRESS;
        }
        else
        {
            dwFlags |= PROGDLG_NOPROGRESSBAR;
        }

        m_ProgressDlg->StartProgressDialog(
            hWndParent, 
            NULL, 
            dwFlags,
            NULL);

        m_DlgShowed = TRUE;

    }


    return TRUE;

}






