/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "PackedArgs.h"
#include <malloc.h>
#include "ChkMAlloc.h"

CPackedArgs::CPackedArgs()
{
}

CPackedArgs::~CPackedArgs()
{
    Empty();
}


void *
CPackedArgs::AllocateMem( ULONG MemSize )
{
    return MALLOC( MemSize );
}


void
CPackedArgs::FreeMem( PVOID pPtr )
{
	FREE( pPtr );
}

size_t
CPackedArgs::GetCount()
{
	return m_Args.size();
}

int
CPackedArgs::GetArgType( UINT ArgNum )
{
	if ( ArgNum >= GetCount() )
		return 0;

	return m_Args[ArgNum].ArgType;
}

bool
CPackedArgs::GetUInt32Arg( UINT ArgNum, UINT *pValue )
{
	if ( GetArgType(ArgNum) != ARGT_UINT32 )
		return false;

	*pValue = m_Args[ArgNum].ArgValue.uArg;
	return true;
}


bool
CPackedArgs::GetUInt64Arg( UINT ArgNum, UINT64 *pValue )
{
	if ( GetArgType(ArgNum) != ARGT_UINT64 )
		return false;

	*pValue = m_Args[ArgNum].ArgValue.uArg64;
	return true;
}

bool
CPackedArgs::GetPtrArg( UINT ArgNum, PVOID *ppPtr, PULONG pcb )
{
	if ( GetArgType(ArgNum) != ARGT_PTR && GetArgType(ArgNum) != ARGT_DYNAMIC_PTR )
		return false;

	*ppPtr = m_Args[ ArgNum ].ArgValue.PtrStruct.pPtr;
	*pcb = m_Args[ ArgNum ].ArgValue.PtrStruct.cbBuffer;

	return true;
}

bool
CPackedArgs::Pack( PVOID *pPtr, PULONG pSize )
{
	UINT BufferSize = 0;

	for (
		vector<ARG_STRUCT>::iterator iter = m_Args.begin();
		iter != m_Args.end();
		iter++
			)
	{
		switch ( iter->ArgType )
		{
			case ARGT_UINT32:
			{
				BufferSize += sizeof( UINT );
				break;
			}

			case ARGT_UINT64:
			{
				BufferSize += sizeof( UINT64 );
				break;
			}

            case ARGT_DYNAMIC_PTR:
			case ARGT_PTR:
			{
				BufferSize += iter->ArgValue.PtrStruct.cbBuffer;
				break;
			}

			default:
			{
				return false;
			}

		}
	}

	BufferSize += sizeof( PACKED_ARGS_HEADER );
	BufferSize += sizeof( PACKED_ARGS_ITEM ) * (UINT)GetCount();

	PUCHAR pBuffer = (PUCHAR)AllocateMem( BufferSize );

	if ( NULL == pBuffer )
		return false;


	PACKED_ARGS_HEADER *pHeader = (PACKED_ARGS_HEADER *)pBuffer;

	pHeader->Count = (UINT)GetCount();
	pHeader->Signature = PACKED_ARGS_SIG;

	PPACKED_ARGS_ITEM pItem = (PPACKED_ARGS_ITEM)((PUCHAR)pHeader + sizeof(PACKED_ARGS_HEADER));


	for (
		vector<ARG_STRUCT>::iterator iter = m_Args.begin();
		iter != m_Args.end();
		iter++
			)
	{
        if ( iter->ArgType == ARGT_DYNAMIC_PTR )
        {
            pItem->ArgType = ARGT_PTR;
        }
        else
        {
		    pItem->ArgType = iter->ArgType;
        }

		switch ( iter->ArgType )
		{
			case ARGT_UINT32:
			{
				pItem->ArgSize = sizeof(UINT);
				PUINT puValue = (PUINT) ( (PUCHAR)pItem +  sizeof( PACKED_ARGS_ITEM ) );

				*puValue = iter->ArgValue.uArg;
				pItem = (PPACKED_ARGS_ITEM)((PUCHAR)puValue + pItem->ArgSize);

				break;
			}

			case ARGT_UINT64:
			{
				pItem->ArgSize = sizeof(UINT64);
				PUINT64 puValue64 = (PUINT64) ( (PUCHAR)pItem +  sizeof( PACKED_ARGS_ITEM ) );

				*puValue64 = iter->ArgValue.uArg64;
				pItem = (PPACKED_ARGS_ITEM)((PUCHAR)puValue64 + pItem->ArgSize);

				break;
			}

            case ARGT_DYNAMIC_PTR:
			case ARGT_PTR:
			{
				pItem->ArgSize = iter->ArgValue.PtrStruct.cbBuffer;

				PUCHAR pValue = (PUCHAR) ( (PUCHAR)pItem +  sizeof( PACKED_ARGS_ITEM ) );
				memcpy( pValue, iter->ArgValue.PtrStruct.pPtr, pItem->ArgSize );

				pItem = (PPACKED_ARGS_ITEM)((PUCHAR)pValue + pItem->ArgSize);
				

				break;
			}

			default:
			{
				FreeMem( pBuffer );
				return false;
			}

		}
	}

	*pPtr = pBuffer;
	*pSize = BufferSize;

	return true;
}


bool
CPackedArgs::Unpack( PVOID pPtr, ULONG BuffSize )
{
	Empty();

	PACKED_ARGS_HEADER *pHeader = (PACKED_ARGS_HEADER *)pPtr;

	if ( BuffSize < sizeof( PACKED_ARGS_HEADER ) ||
		pHeader->Signature != PACKED_ARGS_SIG || 
		sizeof(PACKED_ARGS_HEADER) + pHeader->Count*sizeof(PACKED_ARGS_ITEM) > BuffSize
		)
	{
		return false;
	}

	PUCHAR pEndOfBuffer = (PUCHAR)pPtr + BuffSize;

	PACKED_ARGS_ITEM *pItem = (PPACKED_ARGS_ITEM) ( (PUCHAR)pHeader + sizeof(PACKED_ARGS_HEADER) );

	for ( UINT i = 0; i < pHeader->Count; i++ )
	{
		if  ( 
			(PUCHAR)pItem + sizeof(PACKED_ARGS_ITEM) >= pEndOfBuffer ||
			pItem->ArgSize >= MAX_ITEM_SIZE ||
			(PUCHAR)pItem + pItem->ArgSize >= pEndOfBuffer
			)
		{
			Empty();
			return false;
		}

		switch ( pItem->ArgType )
		{
			case ARGT_UINT32:
			{
				if ( pItem->ArgSize != sizeof(UINT) )
				{
					Empty();
					return false;
				}

				PUINT puValue = (PUINT)((PUCHAR)pItem + sizeof(PACKED_ARGS_ITEM));

				ARG_STRUCT NewArg;

				NewArg.ArgType = pItem->ArgType;
				NewArg.ArgValue.uArg = *puValue;

				m_Args.push_back( NewArg );

				break;
			}


			case ARGT_UINT64:
			{
				if ( pItem->ArgSize != sizeof(UINT64) )
				{
					Empty();
					return false;
				}

				PUINT64 puValue64 = (PUINT64)((PUCHAR)pItem + sizeof(PACKED_ARGS_ITEM));

				ARG_STRUCT NewArg;

				NewArg.ArgType = pItem->ArgType;
				NewArg.ArgValue.uArg64 = *puValue64;

				m_Args.push_back( NewArg );

				break;
			}

			case ARGT_PTR:
            //(!) ARGT_DYNAMIC_PTR is not supported for unpacking
			{
				PVOID pValue = ((PUCHAR)pItem + sizeof(PACKED_ARGS_ITEM));

				ARG_STRUCT NewArg;

				NewArg.ArgType = pItem->ArgType;
				NewArg.ArgValue.PtrStruct.pPtr = pValue;
				NewArg.ArgValue.PtrStruct.cbBuffer = pItem->ArgSize;

				m_Args.push_back( NewArg );


				break;
			}

			default:
			{
				Empty();
				return false;
			}
		}


		pItem = (PACKED_ARGS_ITEM *)( (PUCHAR)pItem + 
			sizeof(PACKED_ARGS_ITEM) +
			pItem->ArgSize);

	} // for


	return true;
}

bool
CPackedArgs::CreateFromArgs1(
	int ArgType1, 
	PVOID pArgValue1
	)
{
	bool bResult = false;

	Empty();

	bResult = PushArg( ArgType1, pArgValue1 );

	if ( !bResult )
	{
		Empty();
		return false;
	}

	return bResult;
}

bool
CPackedArgs::PushArg(
	int ArgType, 
	PVOID pArgValue
	)
{
	ARG_STRUCT NewArg;

	NewArg.ArgType = ArgType;

	switch ( ArgType )
	{
		case ARGT_UINT32:
		{
			PUINT puValue = (PUINT)pArgValue;

			NewArg.ArgValue.uArg = *puValue;
			break;
		}

		case ARGT_UINT64:
		{
			PUINT64 puValue64 = (PUINT64)pArgValue;

			NewArg.ArgValue.uArg64 = *puValue64;
			break;
		}

        case ARGT_DYNAMIC_PTR:
		case ARGT_PTR:
		{
			PPTR_STRUCT pPtrStruct = (PPTR_STRUCT)pArgValue;

			NewArg.ArgValue.PtrStruct.pPtr = pPtrStruct->pPtr;
			NewArg.ArgValue.PtrStruct.cbBuffer = pPtrStruct->cbBuffer;

			break;
		}



		default:
		{
			return false;
			break;
		}


	}

	m_Args.push_back( NewArg );

	return true;	

}




bool
CPackedArgs::CreateFromArgs2(
	int ArgType1, 
	PVOID pArgValue1,
	int ArgType2, 
	PVOID pArgValue2
	)
{
	bool bResult = false;

	Empty();

	bResult = PushArg( ArgType1, pArgValue1 );

	if ( !bResult )
	{
		Empty();
		return false;
	}

	bResult = PushArg( ArgType2, pArgValue2 );

	if ( !bResult )
	{
		Empty();
		return false;
	}


	return bResult;

}

bool
CPackedArgs::CreateFromArgs3(
	int ArgType1, 
	PVOID pArgValue1,
	int ArgType2, 
	PVOID pArgValue2,
	int ArgType3, 
	PVOID pArgValue3
	)
{
	bool bResult = false;

	Empty();

	bResult = PushArg( ArgType1, pArgValue1 );

	if ( !bResult )
	{
		Empty();
		return false;
	}

	bResult = PushArg( ArgType2, pArgValue2 );

	if ( !bResult )
	{
		Empty();
		return false;
	}

	bResult = PushArg( ArgType3, pArgValue3 );

	if ( !bResult )
	{
		Empty();
		return false;
	}


	return bResult;

}


void
CPackedArgs::Empty()
{
    //free dynamic pointers

    for (
        vector<ARG_STRUCT>::iterator iter = m_Args.begin();
        iter != m_Args.end();
        iter++
        )
    {
        if ( iter->ArgType == ARGT_DYNAMIC_PTR )
        {
            FreeMem( iter->ArgValue.PtrStruct.pPtr );
        }
    }

	m_Args.clear();
}
